<?php
// pfSense log file scanner
// send email on ipsec VPN connect, admin login or other log file entry

// Setup:
// 1. set config in logScan.json - put desired logs to monitor into json file, set email address, etc.
//
// 2. set Status | System Logs | Settings to RFC 5424
//
// 3. copy logScanRB8r.php and logScan.json to /root (Diagnostics | Edit File)
// change permissions (Diagnostics | Command Prompt | Execute Shell Command):
//   cd /root; chmod 600 logScan*; ls -l logScan*
// 
// 4. run from command line to setup (Diagnostics | Command Prompt | Execute Shell Command): 
//    cd /root; php logScanRB8r -i
// if no errors, then
// 
// 5. Modify cron entry: (Services | Cron | edit pencil)
// find entry: /usr/sbin/newsyslog
// change to: php /root/logScanRB8.php >> /root/logScan.log 2>&1; /usr/sbin/newsyslog
// -may want to change to every 2 minutes
//
// Note: if you set $debug=true, you should probably run "cd /root; chmod 600 logScan*; ls -l logScan*" again
// so that logScan.log is restricted to the root user, as your email passwords may be there.
//
// There are two variables in the .json file that are updated each run:
// time - the unix epoch time of the start of the log file (to sense log rotation)
// offset - byte offset into the log file of text that has already been scanned (to avoid duplicate alerts)
//
// MIT License:
// Copyright 2023 Robert Bleidt
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
// documentation files (the “Software”), to deal in the Software without restriction, including without 
// limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the
// Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all copies or substantial portions 
// of the Software.
//
// THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
// TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
// THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
// CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
// DEALINGS IN THE SOFTWARE.
 
require_once 'Mail.php'; 
$debug = false;
$emailDebug = false;
$initialize = false;
if ($debug) print "debug = ".$debug."\n";

date_default_timezone_set('UTC');

 if (strpos($argv[1],'-i') !== false) {
	$initialize = true;
	print "Initialize Mode\n";
}

// read in the config for files to watch
if (($jsonData = file_get_contents('logScan.json',false,null,0,10000)) === false) {sendFatal("JSON file read error\n"); }
$config = json_decode($jsonData,true, 10) or sendFatal("JSON decoded error\n");
if ($debug) var_dump($config);

$hostName = $config["configuration"]["hostname"];

foreach ($config["logs"] as $key => $log) {
	if ($debug) print date("r"). " processing ".$config["logs"][$key]["file"]."\n";

	$curLogFileName = $config["logs"][$key]["file"]; //current log 
	//$priorLogFileName = $config["logs"][$key]["priorFile"]; // prior logs may be bzip2 compressed, check setting

	$scanTime = intval($config["logs"][$key]["time"]);
	$scanPos = intval($config["logs"][$key]["offset"]);
	if ($debug) print date("r")." scanTime: $scanTime  scanPos: $scanPos\n";

	// get the time of the first log entry
	$logStartTime = getLogFileStartTime($curLogFileName);
	if ($debug) print date("r"). " LogScanRB log start time: $logStartTime\n";
	if ($logStartTime == 0) {
		// xxx add counter here, store in json?
		if ($debug) print date("r"). " LogScanRB log start time is zero: $logStartTime\n";
		continue;
	}

	if ($initialize) {
		$scanPos = 0;
		$scanTime = $logStartTime;
	}
	else {
		if ($logStartTime > $scanTime) {
			// logfile has rolled over
			print date("r")." LogScanRB Alert: $curLogFileName rollover\nlogStartTime: $logStartTime  scanTime: $scanTime\n";
			$scanPos = 0; // reset as now the new file will be scanned
			$scanTime = $logStartTime;
		}
	}
	scanLogFile($curLogFileName, $scanPos, $config["logs"][$key]["strings"],$initialize);

	// write the unix time of the last log entry scanned back to the text file, for next run.
	$config["logs"][$key]["time"] = $scanTime;
	$config["logs"][$key]["offset"] = $scanPos;
	
	$fp = fopen('logScan.json', 'w') or sendFatal("unable to open JSON file for writing");
	fwrite($fp, json_encode($config,JSON_PRETTY_PRINT)) or sendFatal("JSON file write error");
	fclose($fp) or sendFatal("JSON file close error");;

	if ($debug) print date("r")." updated scanTime: $scanTime  scanPos: $scanPos\n";
}

exit(0);


 function getLogFileStartTime ($fileName) {
	// get the unix time of the first log entry in the file by reading first line of text in the log.
	// Complicated since log might be empty  due to log rotation or have a non-RFC 5424 date.
	// Looking at the file time probably won't work, since mtime is updated on each entry. 
	// not sure of ctime on FreeBSD
	//
	// if the time can't be determined, returns 0.
	
	global $debug;
	$logFile = fopen($fileName, "r") or sendFatal("getLogFileStartTime: Unable to open log file: $fileName"); 
	if ($logFileLine = fgets($logFile)) {
		if ($logFileLine[0] != '<') {
			// take care of extra line from syslog at top of file? RFC 5424 says line should start with "<"
			// ipsec.log has old-style date on rollover message
			if ($debug) print date("r")." getLogFileStartTime: skipping first line\n";
			if ($logFileLine = fgets($logFile)) {
				if ($debug) print date("r")." getLogFileStartTime: using second line $logFileLine\n";
			}
			else {
				if ($debug) print date("r")." getLogFileStartTime: first line skipped, nothing else\n";
				fclose($logFile) or sendFatal("getLogFileStartTime: log file $fileName close error");
				return (0);
			}
		}
		if ($logFileLine[0] != '<') {
			sendFatal("getLogFileStartTime: log file $fileName read error on second line");
		}
		fclose($logFile) or sendFatal("getLogFileStartTime: log file $fileName close error");
		
		// line read, now process it.
		if ($debug) print date("r")." getLogFileStartTime: line read: $logFileLine\n";
		$logFileLineArray = explode(" ",$logFileLine,3);
		if ($debug) print date("r")." getLogFileStartTime: array: $logFileLineArray[1]\n";
		//var_dump ($grepArray);
		if  (! ($foundTime = strtotime($logFileLineArray[1]))) {
			if ($debug) print date("r")." getLogFileStartTime: strtotime returned false: $logFileLineArray[1]\n";
			sendFatal("getLogFileStartTime: strtotime returned false: $fileName");
		}
		return ($foundTime);
	}
	fclose($logFile) or sendFatal("getLogFileStartTime: log file  $fileName close error");
	if ($debug) print date("r")." getLogFileStartTime: file is empty: $fileName\n";
	// sendFatal("getLogFileStartTime: file is empty: $fileName");
	return (0);
}

function scanLogFile ($fileName, &$scanPos, $strings, $initialize) {
	// scan log file $fileName and sendAlert on any lines with a string match
	global $debug;
	$logFile = fopen($fileName, "r") or sendFatal("scanLogFile: Unable to open log file $fileName"); 
	if (fseek ($logFile, $scanPos) != 0) sendFatal("scanLogFile: Unable to fseek to $scanpos in $fileName"); 
	while($logFileLine = fgets($logFile)) {
		if (!$initialize) foreach($strings as $str) {
			if (stripos($logFileLine,$str) !== false) {
					print date("r")." LogScanRB Alert: $logFileLine \n";
					// Shift off priority and time (to save SMS space) and send alert
					$logFileLineArray = explode(" ",$logFileLine);
					array_shift($logFileLineArray);
					array_shift($logFileLineArray);
					$logFileLine = implode(" ",$logFileLineArray);
					sendAlert($logFileLine);
					break;
			}	
		}
	}
	$scanPos = ftell($logFile);
	if ($debug) print date("r")." scanLogFile: finished scanPos: $scanPos\n";
	fclose($logFile) or sendFatal("scanLogFile: log file $fileName close error");
}

Function sendFatal ($str) {
  print "logScanRB: Fatal Error: ".$str."\n";
	sendAlert ($str);
	exit (1);
}

function sendAlert ($str) {
	global $config;
	//$text =  "LogScanRB Alert: " .date("r") . $str ;
	$text =  date("M-d H:i:s") ." ". $str ;
	foreach ($config["configuration"]["mailTo"] as $key => $value) {
		sendMail($config["configuration"]["mailTo"][$key]["mx"],
			$config["configuration"]["mailTo"][$key]["port"],
			$config["configuration"]["mailTo"][$key]["to"],
			"LogScanRB", $text,
			$config["configuration"]["mailTo"][$key]["useAuth"],
			$config["configuration"]["mailTo"][$key]["user"],
			$config["configuration"]["mailTo"][$key]["pw"]		);
	}
}

function sendMail($mxHost,$port,$to,$subject,$message,$useAuth,$user,$pw) {
	global $hostName, $emailDebug;
	print date("r")." sendMail: $mxHost to $to\n";
	$params = array(
		'host' => $mxHost, // needs to be MX server name
		'port' => $port,
		'auth' => (boolean) $useAuth,
		'username' => $user,
		'password' => $pw,
		'localhost' => $hostName,
		'timeout' => 20,
		'debug' => $emailDebug,
		'persist' => false
	);
	
	$headers = array(
		"From"    => $hostName,
		"To"      => $to,
		"Subject" => $subject,
		"Date"    => date("r")
	);
	
	try {
		$smtp =& Mail::factory('smtp', $params);
		$mail = @$smtp->send($to, $headers, $message);
		if (PEAR::isError($mail)) {
			print date("r")." PEAR Error: ".gettext($error_text). " To:". $to. "Message:". $mail->getMessage()."\n";
		}
	} catch (Exception $e) {
			print date("r")." Mail Error: ". " To:". $to. "Message:". $e->getMessage()."\n";
	}
}


?> 
